use std::ffi::CStr;

use libc::{setlocale, sigaction, sigemptyset, sigset_t, waitpid, LC_CTYPE, SA_NOCLDSTOP, SA_NOCLDWAIT, SA_RESTART, SIGCHLD, SIG_IGN, WNOHANG};
use x11::xlib::{SubstructureRedirectMask, XDefaultRootWindow, XDefaultScreen, XDisplayHeight, XDisplayWidth, XErrorEvent, XOpenDisplay, XRootWindow, XSelectInput, XSetErrorHandler, XSupportsLocale, XSync, _XDisplay};

mod drw;

// TODO: Replace with actual configs
const FONT: &'static str = "monospace:size=10";

#[derive(Debug)]
pub struct Server {
    display: *mut _XDisplay,
    screen: Screen,
    error_handler: Option<ErrorHandler>,
    root: u64,
    drw: Option<drw::Drw>
}

#[derive(Debug, Default)]
pub struct Screen {
    sn: i32,
    w: i32,
    h: i32,
}

type ErrorHandler = unsafe extern "C" fn (display: *mut _XDisplay, ev: *mut XErrorEvent) -> i32;

impl Server {
    pub fn new() -> Self { 
        unsafe {
            Self { display: std::ptr::null_mut(), screen: Default::default(), error_handler: None, root: 0, drw: None }
        }
    }

    pub fn setup(&mut self) {
        //sets locale and tests if other wms are runnnig
        unsafe {
            if setlocale(LC_CTYPE, c"".as_ptr()).is_null() || XSupportsLocale() != 0 {
                log::warn!("No locale support");
            }
            self.display = XOpenDisplay(std::ptr::null());
            
            self.error_handler = XSetErrorHandler(Some(Self::error_wm_exists));
            /* this causes an error if some other window manager is running */
            let drw = XDefaultRootWindow(self.display);
            XSelectInput(self.display, drw, SubstructureRedirectMask);
            XSync(self.display, false as i32);
            XSetErrorHandler(Some(Self::error_handler));
            XSync(self.display, false as i32);
        }
        // do not transform children into zombies when they terminate
        unsafe {
            let mut sa: sigaction = std::mem::zeroed();
            sigemptyset(&mut sa.sa_mask as *mut sigset_t);
            sa.sa_flags = SA_NOCLDSTOP | SA_NOCLDWAIT | SA_RESTART;
            sa.sa_sigaction = SIG_IGN;
            sigaction(SIGCHLD, &sa, std::ptr::null_mut());
        }

        // clean up any zombies (inherited from .xinitrc etc) immediately
        unsafe {
    	    while waitpid(-1, std::ptr::null_mut(), WNOHANG) > 0 {};
        }
        // Init screen
        unsafe {
            let s = XDefaultScreen(self.display);
            self.screen.w = XDisplayWidth(self.display, s);
            self.screen.h = XDisplayHeight(self.display, s);
            self.root = XRootWindow(self.display, s);
            self.drw = Some(drw::Drw::new(self.display, self.screen.sn, self.root, self.screen.w, self.screen.h, vec![FONT]).unwrap());

        }
    }

    fn drw(&mut self) -> drw::Drw {
        &a
    }

    unsafe extern "C" fn error_wm_exists(_display: *mut _XDisplay, _ev: *mut XErrorEvent) -> i32 {
        log::error!("A WM is already running");
        return -1; 
    }

    unsafe extern "C" fn error_handler(_display: *mut _XDisplay, _ev: *mut XErrorEvent) -> i32 {
        log::error!("idk what happened good luck lmao");
        return -1; 
    } 

}
